#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <sys/kernel.h>
#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <X11/Xutil.h>

Display        *display;
int             screen;
Colormap        colormap;
char           *progname;
FILE           *dumpfp = NULL;
int             pc_colour[16];
static char *pc_names[] = {
 "Black", "mediumBlue", "Green3", "Cyan3", "Red", "Magenta4", "sienna", "Grey",
 "RoyalBlue1", "blue", "green", "cyan", "tomato", "magenta", "yellow", "white"
};

get_colour(char *cname, XColor *cp)
{
	int status = XParseColor(display, colormap, cname, cp) &&
		XAllocColor(display, colormap, cp);
	if (dumpfp != NULL) {
		char cdef[4];
		cdef[0] = cp->blue & 0xff;
		cdef[1] = cp->green & 0xff;
		cdef[2] = cp->red & 0xff;
		cdef[3] = 0;
		fwrite(cdef, sizeof cdef[0], sizeof cdef/sizeof cdef[0], dumpfp);
	}
	return status;
}

PcColours(void)
{
	int	i;
	XColor  cdef;

	for (i=0; i < 16; i++) {
		if (!get_colour(pc_names[i], &cdef)) return 0;
		pc_colour[i] = cdef.pixel;
	}
	return 1;
}

int
load_named_colours(const char *filename)
{
	FILE *fp;
	char name[LINE_MAX];
	int c, pos = 0, ncells = 0, overrun = 0;

	if (strcmp(filename, "-") == 0) fp = stdin;
	else fp = fopen(filename, "r");

	if (fp == NULL) {
		fprintf(stderr, "%s: couldn't open file %s\n",
			progname, filename);
		return 0;
	}

	while ((c = fgetc(fp)) != -1) {
		XColor  cdef;

		if (isspace(c)) {
			name[pos] = '\0';
			if (overrun) {
				fprintf(stderr, "%s: colour name too long\n",
					progname);
			} else if (pos > 0) {
				if (!get_colour(name, &cdef)) {
					if (ncells == 0) {
						fprintf(stderr, "%s: couldn't load any colormap entries\n",
							progname);
						return 0;
					} else {
						fprintf(stderr, "%s: could only load %d colormap entries\n",
							progname, ncells);
						break;
					}
				}
				ncells++;
			}
			pos = 0;
			overrun = 0;
		} else {
			if (pos < (LINE_MAX-1)) 
				name[pos++] = c;
			else
				overrun = 1;
		}
	}

	return 1;
}

int
load_map_colours(const char *rgbmap)
{
	Atom map = 0;
	XStandardColormap *best_map_info;
	int count;

	if (strcmp(rgbmap, "best") == 0) 
		map = XA_RGB_BEST_MAP;
	else if (strcmp(rgbmap, "default") == 0) 
		map = XA_RGB_DEFAULT_MAP;
	else if (strcmp(rgbmap, "red") == 0) 
		map = XA_RGB_RED_MAP;
	else if (strcmp(rgbmap, "green") == 0) 
		map = XA_RGB_GREEN_MAP;
	else if (strcmp(rgbmap, "blue") == 0) 
		map = XA_RGB_BLUE_MAP;
	else if (strcmp(rgbmap, "gray") == 0) 
		map = XA_RGB_GRAY_MAP;

	if (map == 0) return 0;

	if (XGetRGBColormaps(display, RootWindow(display, screen),
		&best_map_info, &count, map) == 0) {
		fprintf(stderr, "%s: standard %s colormap property isn't set\n",
			progname, rgbmap);
		return 0;
	} else if (best_map_info->colormap) {
		if (best_map_info->red_max == 0) {
			fprintf(stderr, "%s: standard colormap property is set,"
				" but is missing data\n", progname);
			return 0;
		} else {
			int i, j;
			int ncells = best_map_info->base_pixel +
				((best_map_info->red_max + 1) *
				 (best_map_info->green_max + 1) *
				 (best_map_info->blue_max + 1));
			XColor *exact_defs = calloc(sizeof(XColor), ncells);

			if (!exact_defs) {
				fprintf(stderr, "%s: out of memory\n", progname);
				return 0;
			}

			for (i = 0; i < ncells; i++)
				exact_defs[i].pixel = i;

			XQueryColors(display, best_map_info->colormap,
					exact_defs, ncells);
			for (i = 0; i < ncells; i++) {
				if (!XAllocColor(display,
						DefaultColormap(display, screen),
						&exact_defs[i])
					&& j == i) {
					fprintf(stderr, "%s: could only load %d of %d colormap entries\n",
							progname, j, ncells);
					if (dumpfp == NULL)
						break;
				} else
					j++;
				if (dumpfp != NULL) {
					char cdef[4];
					cdef[0] = exact_defs[i].blue & 0xff;
					cdef[1] = exact_defs[i].green & 0xff;
					cdef[2] = exact_defs[i].red & 0xff;
					cdef[3] = 0;
					fwrite(cdef, sizeof cdef[0], sizeof cdef/sizeof cdef[0], dumpfp);
				}
			}
			free(exact_defs);
		}
	} else if (best_map_info->visualid == 0) {
		fprintf(stderr, "%s: standard colormap property is set,"
			" but is missing data\n", progname);
		return 0;
	} else {
		int i, j, k, l, n;
		int ncells = best_map_info->base_pixel +
			((best_map_info->red_max + 1) *
			 (best_map_info->green_max + 1) *
			 (best_map_info->blue_max + 1));
		XColor def;

		n = l = 0;
		for (i = 0; i < best_map_info->blue_max; i++) {
			for (j = 0; j < best_map_info->green_max; j++) {
				for (k = 0; k < best_map_info->red_max; k++) {
					def.red = 0xffff * k / best_map_info->red_max;
					def.green = 0xffff * j / best_map_info->green_max;
					def.blue = 0xffff * i / best_map_info->blue_max;
					if (!XAllocColor(display,
						DefaultColormap(display, screen), &def)
						&& l == n) {
						fprintf(stderr,
							"%s: could only load %d of %d colormap entries\n",
							progname, l, ncells);
						if (dumpfp == NULL)
							break;
					} else 
						l++;
					if (dumpfp != NULL) {
						char cdef[4];
						cdef[0] = def.blue & 0xff;
						cdef[1] = def.green & 0xff;
						cdef[2] = def.red & 0xff;
						cdef[3] = 0;
						fwrite(cdef, sizeof cdef[0], sizeof cdef/sizeof cdef[0], dumpfp);
					}
					n++;
				}
			}
		}
	}

	return 1;
}

main(int argc, char *argv[]) 
{
	int halt = 0;	/* don't do infinite loop */
	int load_pc_colours = 1;
	char *filename = NULL, *rgbmap = NULL, *dumpfile = NULL;
	const char *arg;

	progname = argv[0]; 
	if (argc > 0) {
		while ((arg = *++argv) != NULL) {
			if (strcmp(arg, "-pc") == 0) {
				load_pc_colours = 1;
				filename = NULL;
				rgbmap = NULL;
			} else if (strcmp(arg, "-palette") == 0) {
				load_pc_colours = 0;
				filename = *++argv;
				rgbmap = NULL;
				if (filename == NULL) {
					fprintf(stderr, "usage: %s [-pc] [-palette filename] [-rgbmap map]\n",
						progname);
					exit(1);
				}
			} else if (strcmp(arg, "-rgbmap") == 0) {
				load_pc_colours = 0;
				filename = NULL;
				rgbmap = *++argv;
				if (rgbmap == NULL) {
					fprintf(stderr, "usage: %s [-pc] [-palette filename] [-rgbmap map]\n",
						progname);
					exit(1);
				}
			} else if (strcmp(arg, "-dump") == 0) {
				dumpfile = *++argv;
				if (dumpfile == NULL) {
					fprintf(stderr, "usage: %s [-pc] [-palette filename] [-rgbmap map]\n",
						progname);
					exit(1);
				}
			} else if (strcmp(arg, "-halt") == 0) {
				halt = 1;
			} else {
				fprintf(stderr, "usage: %s [-pc] [-palette filename] [-rgbmap map]\n",
					progname);
				exit(1);
			}
		}
	}

	if (dumpfile != NULL) dumpfp = fopen(dumpfile, "w");

	if (!(display = XOpenDisplay(NULL))) exit(1);
	screen = DefaultScreen(display);
	colormap = DefaultColormap(display,screen);
	if (load_pc_colours && !PcColours()) exit(1);
	else if (filename != NULL && !load_named_colours(filename)) exit(1);
	else if (rgbmap != NULL && !load_map_colours(rgbmap)) exit(1);

	if (dumpfp != NULL) fclose(dumpfp);
	if (halt) exit(0);

	if (fork()) exit(0);
	close(0);close(1);close(2);
	Receive(0,NULL,0);
}

